
#include <stdbool.h>
#include <pthread.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <uci.h>
#include <string.h>
#include <time.h>

#include "host_config.h"
#include "topologyd.h"
#include "host.h"
#include "syslog.h"
#include "host_wan_stats.h"

#define stats_rd_rx_socket  priv->host_stats_rd_sockets[0]
#define stats_update_rx_socket  priv->host_stats_update_sockets[0]
#define stats_update_wr_socket  priv->host_stats_update_sockets[1]

static void host_wan_stats_update_per_node(struct host_node *host, struct host_nf_tc_ip_list *n)
{
	host->stats.tx_packets += n->tx_packets;
	host->stats.tx_bytes += n->tx_bytes;
	host->stats.rx_packets += n->rx_packets;
	host->stats.rx_bytes += n->rx_bytes;
	host->active_connections = n->ct_running_count;
	n->tx_packets = 0;
	n->tx_bytes = 0;
	n->rx_packets = 0;
	n->rx_bytes = 0;
	n->ct_running_count = 0;
	n->in_use = HOST_WAN_STATS_IP_NODE_EXPIRED;
}

static void host_wan_stats_node_traverse(struct topologyd_private *priv, struct host_node *host)
{
	int count;
	struct host_nf_tc_ip_list *n = NULL;
	struct sockaddr_in sa;
	struct sockaddr_in6 sa6;

	for (count=0; count < host->ipv4addr_count; count++) {
		if (inet_pton(AF_INET, host->ipv4addr_list[count], &(sa.sin_addr)) <= 0)
			continue;
		hlist_for_each_entry(n, &priv->host.ct_ip_node_htable[ct_ip_node_hash(ntohl(sa.sin_addr.s_addr))], hlist) {
			if (ntohl(sa.sin_addr.s_addr) == n->ipv4_addr.s_addr)
			{
				host_wan_stats_update_per_node(host, n);
				break;
			}
		}
	}
	for (count=0; count < host->ipv6addr_count; count++) {
		if (inet_pton(AF_INET6, host->ipv6addr[count], &sa6.sin6_addr) <= 0)
			continue;
		hlist_for_each_entry(n, &priv->host.ct_ipv6_node_htable[ct_ip_node_hash(sa6.sin6_addr.s6_addr[15])], hlist) {
			if (!memcmp(n->ipv6_addr.sin6_addr.s6_addr, sa6.sin6_addr.s6_addr, sizeof(int)*4)) {
				host_wan_stats_update_per_node(host, n);
				break;
			}
		}
	}
}

static void host_wan_stats_update(struct topologyd_private *priv)
{
	int i;
	struct host_node *host = NULL;
	struct device_address_list *addr_list_n = NULL;
	struct device_address_list *temp_n = NULL;

	for (i = 0; i < NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.node_htable[i])))
			continue;

                hlist_for_each_entry(host, &priv->host.node_htable[i], hlist) {
                        if (hwaddr_is_zero(host->hwaddr) || (host->is_copy == 1) ||
				(host->intf_type != HOST_TYPE_ETHER && host->intf_type != HOST_TYPE_WIFI))
                                continue;

			host->active_connections = 0;
			host_wan_stats_node_traverse(priv, host);
		}
	}
	if (hlist_empty(&priv->device_addr_list_table))
		return;
	hlist_for_each_entry(addr_list_n, &priv->device_addr_list_table, hlist) {
		if (temp_n) {
			hlist_del(&temp_n->hlist, &priv->device_addr_list_table);
			free(temp_n);
		}
		temp_n = addr_list_n;
	}
	if (temp_n) {
		hlist_del(&temp_n->hlist, &priv->device_addr_list_table);
		free(temp_n);
	}
}

int host_wan_stats_thread(struct topologyd_private *priv)
{
    int err;
    pthread_t thread;
    pthread_attr_t attr;

    // Exit if unable to create thread attributes
    err = pthread_attr_init(&attr);
    if (err != 0) {
        syslog(LOG_ERR, "pthread_attr_init failed err:%d", err);
        return -1;
    }

    // Exit if unable to create the thread as detached (as we do not need to wait for it to terminate)
    err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if (err != 0) {
        syslog(LOG_ERR, "pthread_attr_setdetachstate failed err:%d", err);
        goto exit;
    }

    // Exit if unable to start a new thread to perform the operation
    err = pthread_create(&thread, &attr, host_wan_stats, priv);
    if (err != 0) {
        syslog(LOG_ERR, "pthread_create failed err:%d", err);
        goto exit;
    }
    err = pthread_setname_np(thread, "topology_pm");
    if (err != 0) {
        syslog(LOG_INFO, "pthread set name failed err:%d", err);
    }

exit:
    pthread_attr_destroy(&attr);
    return err;
}

void host_wan_stats_timer_callback(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, host_wan_stats_timer);
	int stats_rd_wr_socket = priv->host_stats_rd_sockets[1];
	host_prepare_device_address_list(priv, AF_INET);
	host_prepare_device_address_list(priv, AF_INET6);

	host_wan_topology_pm_update(stats_rd_wr_socket, priv);

}

void host_wan_stats_collect_timer(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, host_wan_stats_timer);
	priv->host_wan_stats_timer.cb = host_wan_stats_timer_callback;
	uloop_timeout_set(&priv->host_wan_stats_timer, 10 * 1000);
}

void host_wan_stats_read_callback(struct uloop_fd *u, unsigned int events)
{
	void *msg;
	int bytes_read;
	struct topologyd_private *priv;
	bytes_read = recv(u->fd, &msg, sizeof(void *), 0);
	priv = (struct topologyd_private *)msg;
	if (bytes_read != sizeof(msg))
		syslog(LOG_INFO, "%s %d bytes_read: %d priv:%p", __FUNCTION__, __LINE__, bytes_read,priv);
	host_wan_stats_update(priv);
	host_wan_stats_collect_timer(&priv->host_wan_stats_timer);
}

void host_wan_stats_fd_register(struct topologyd_private *priv)
{
                                                                               
        memset(&priv->host_stats_uloop_fd, 0, sizeof(struct uloop_fd));                            
        priv->host_stats_uloop_fd.cb = host_wan_stats_read_callback;                                          
        priv->host_stats_uloop_fd.fd = stats_update_rx_socket;                                               
        if(uloop_fd_add(&priv->host_stats_uloop_fd, ULOOP_READ | ULOOP_EDGE_TRIGGER)) {
		syslog(LOG_ERR, "%s %d uloop_fd add failed", __FUNCTION__, __LINE__);
	}
}

void host_node_create_rd_wr_sockets(struct topologyd_private *priv)
{
	int err;
	err = socketpair(AF_UNIX, SOCK_DGRAM, 0, priv->host_stats_rd_sockets);
	if (err != 0) {
		syslog(LOG_ERR,"failed to open socket topology_stats_rx_sockets err:%d", err);
		return;
	}
	err = socketpair(AF_UNIX, SOCK_DGRAM, 0, priv->host_stats_update_sockets);
	if (err != 0) {
		syslog(LOG_ERR,"failed to open socket topology_stats_tx_sockets err:%d", err);
	}
}

void host_wan_topology_pm_update(int socket_fd, struct topologyd_private *priv)
{
    void *msg = (void *) priv;
    int bytes_sent;

    // Send the message
    bytes_sent = send(socket_fd, &msg, sizeof(void *), MSG_DONTWAIT);
    if (bytes_sent != sizeof(msg)) {
        syslog(LOG_ERR,"%s(%d): send failed : (err=%d)", __FUNCTION__, __LINE__, errno );
    }
}


static void host_stats_select_fds(int sock)
{
    int num_sockets;
    fd_set readfds;
    int numfds;
    void *msg;
    struct topologyd_private *priv = NULL;
    struct timeval timeout;
    int bytes_read;

    timeout.tv_sec = 10;
    timeout.tv_usec = 0;
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    numfds=sock;
    num_sockets = select(numfds+1, &readfds, NULL, NULL, &timeout);
    if (num_sockets > 0) {
	if( FD_ISSET(sock, &readfds) == 0)
		return;

	bytes_read = recv(sock, &msg, sizeof(void *), 0);
	if (bytes_read == sizeof(msg)) {
    		priv = (struct topologyd_private *)msg;

		host_node_get_statistics(priv);
		host_wan_topology_pm_update(stats_update_wr_socket, priv);
	}
    }
}

void *host_wan_stats(void *pData)
{
	struct topologyd_private *priv = (struct topologyd_private *)pData;
	while(1) {
		host_stats_select_fds(stats_rd_rx_socket);
	}
}
